==> google-cloud-cpp-0.9.0.tar.gz: google-cloud-cpp-0.9.0/google/cloud/storage/internal/curl_resumable_streambuf.cc <==
// Copyright 2018 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "google/cloud/storage/internal/curl_resumable_streambuf.h"
#include "google/cloud/log.h"

namespace google {
namespace cloud {
namespace storage {
inline namespace STORAGE_CLIENT_NS {
namespace internal {

CurlResumableStreambuf::CurlResumableStreambuf(
    std::unique_ptr<ResumableUploadSession> upload_session,
    std::size_t max_buffer_size, std::unique_ptr<HashValidator> hash_validator)
    : upload_session_(std::move(upload_session)),
      max_buffer_size_(UploadChunkRequest::RoundUpToQuantum(max_buffer_size)),
      hash_validator_(std::move(hash_validator)),
      last_response_{400} {
  current_ios_buffer_.reserve(max_buffer_size_);
}

bool CurlResumableStreambuf::IsOpen() const {
  return static_cast<bool>(upload_session_);
}

bool CurlResumableStreambuf::ValidateHash(ObjectMetadata const& meta) {
  hash_validator_->ProcessMetadata(meta);
  hash_validator_result_ = std::move(*hash_validator_).Finish();
  return !hash_validator_result_.is_mismatch;
}

CurlResumableStreambuf::int_type CurlResumableStreambuf::overflow(int_type ch) {
  if (!IsOpen()) {
    return traits_type::eof();
  }
  auto status = Flush(false);
  if (!status.ok()) {
    return traits_type::eof();
  }
  if (!traits_type::eq_int_type(ch, traits_type::eof())) {
    current_ios_buffer_.push_back(traits_type::to_char_type(ch));
    pbump(1);
  }
  return 0;
}

int CurlResumableStreambuf::sync() {
  auto status = Flush(false);
  if (!status.ok()) {
    return traits_type::eof();
  }
  return 0;
}

std::streamsize CurlResumableStreambuf::xsputn(char const* s,
                                               std::streamsize count) {
  if (!IsOpen()) {
    return traits_type::eof();
  }
  auto status = Flush(false);
  if (!status.ok()) {
    return traits_type::eof();
  }

  current_ios_buffer_.append(s, static_cast<std::size_t>(count));
  pbump(static_cast<int>(count));
  return count;
}

StatusOr<HttpResponse> CurlResumableStreambuf::DoClose() {
  GCP_LOG(INFO) << __func__ << "()";
  return Flush(true);
}

StatusOr<HttpResponse> CurlResumableStreambuf::Flush(bool final_chunk) {
  if (!IsOpen()) {
    return last_response_;
  }
  // Shorten the buffer to the actual used size.
  auto actual_size = static_cast<std::size_t>(pptr() - pbase());
  if (actual_size == 0) {
    return last_response_;
  }
  if (actual_size <= max_buffer_size_ && !final_chunk) {
    return last_response_;
  }

  std::string trailing;
  std::size_t upload_size = 0U;
  if (final_chunk) {
    current_ios_buffer_.resize(actual_size);
    upload_size =
        upload_session_->next_expected_byte() + current_ios_buffer_.size();
  } else {
    trailing = current_ios_buffer_.substr(max_buffer_size_);
    current_ios_buffer_.resize(max_buffer_size_);
  }
  hash_validator_->Update(current_ios_buffer_);

  auto result = upload_session_->UploadChunk(current_ios_buffer_, upload_size);
  if (!result.ok()) {
    // This was an unrecoverable error, time to signal an error.
    return std::move(result).status();
  }
  current_ios_buffer_.clear();
  current_ios_buffer_.reserve(max_buffer_size_);
  setp(&current_ios_buffer_[0], &current_ios_buffer_[0] + max_buffer_size_);
  current_ios_buffer_.append(trailing);
  pbump(static_cast<int>(trailing.size()));

  if (final_chunk) {
    upload_session_.reset();
  }

  // If `result.ok() == false` we never get to this point, so the last response
  // was actually successful, represent that by a HTTP 200 status code.
  last_response_ = HttpResponse{200, std::move(result).value().payload, {}};
  return last_response_;
}

}  // namespace internal
}  // namespace STORAGE_CLIENT_NS
}  // namespace storage
}  // namespace cloud
}  // namespace google

==> google-cloud-cpp-0.10.0.tar.gz: google-cloud-cpp-0.10.0/google/cloud/storage/internal/curl_resumable_streambuf.cc <==
// Copyright 2018 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "google/cloud/storage/internal/curl_resumable_streambuf.h"
#include "google/cloud/log.h"

namespace google {
namespace cloud {
namespace storage {
inline namespace STORAGE_CLIENT_NS {
namespace internal {

CurlResumableStreambuf::CurlResumableStreambuf(
    std::unique_ptr<ResumableUploadSession> upload_session,
    std::size_t max_buffer_size, std::unique_ptr<HashValidator> hash_validator)
    : upload_session_(std::move(upload_session)),
      max_buffer_size_(UploadChunkRequest::RoundUpToQuantum(max_buffer_size)),
      hash_validator_(std::move(hash_validator)),
      last_response_{400} {
  current_ios_buffer_.reserve(max_buffer_size_);
  auto pbeg = &current_ios_buffer_[0];
  auto pend = pbeg + current_ios_buffer_.size();
  setp(pbeg, pend);
}

bool CurlResumableStreambuf::IsOpen() const {
  return static_cast<bool>(upload_session_);
}

bool CurlResumableStreambuf::ValidateHash(ObjectMetadata const& meta) {
  hash_validator_->ProcessMetadata(meta);
  hash_validator_result_ = std::move(*hash_validator_).Finish();
  return !hash_validator_result_.is_mismatch;
}

CurlResumableStreambuf::int_type CurlResumableStreambuf::overflow(int_type ch) {
  if (!IsOpen()) {
    return traits_type::eof();
  }
  if (traits_type::eq_int_type(ch, traits_type::eof())) {
    // For ch == EOF this function must do nothing and return any value != EOF.
    return 0;
  }
  // If the buffer is full flush it immediately.
  auto status = Flush(false);
  if (!status.ok()) {
    return traits_type::eof();
  }
  // Push the character into the current buffer.
  current_ios_buffer_.push_back(traits_type::to_char_type(ch));
  pbump(1);
  return ch;
}

int CurlResumableStreambuf::sync() {
  auto status = Flush(false);
  if (!status.ok()) {
    return traits_type::eof();
  }
  return 0;
}

std::streamsize CurlResumableStreambuf::xsputn(char const* s,
                                               std::streamsize count) {
  if (!IsOpen()) {
    return traits_type::eof();
  }
  current_ios_buffer_.assign(pbase(), pptr());
  current_ios_buffer_.append(s, static_cast<std::size_t>(count));
  auto pbeg = &current_ios_buffer_[0];
  auto pend = pbeg + current_ios_buffer_.size();
  setp(pbeg, pend);
  pbump(static_cast<int>(current_ios_buffer_.size()));

  auto status = Flush(false);
  if (!status.ok()) {
    return traits_type::eof();
  }
  return count;
}

StatusOr<HttpResponse> CurlResumableStreambuf::DoClose() {
  GCP_LOG(INFO) << __func__ << "()";
  return Flush(true);
}

StatusOr<HttpResponse> CurlResumableStreambuf::Flush(bool final_chunk) {
  if (!IsOpen()) {
    return last_response_;
  }
  // Shorten the buffer to the actual used size.
  auto actual_size = static_cast<std::size_t>(pptr() - pbase());
  if (actual_size < max_buffer_size_ && !final_chunk) {
    return last_response_;
  }

  std::string trailing;
  std::size_t upload_size = 0U;
  if (final_chunk) {
    current_ios_buffer_.resize(actual_size);
    upload_size = upload_session_->next_expected_byte() + actual_size;
  } else {
    auto chunk_count = actual_size / UploadChunkRequest::kChunkSizeQuantum;
    auto chunk_size = chunk_count * UploadChunkRequest::kChunkSizeQuantum;
    trailing.assign(pbase() + chunk_size, pbase() + actual_size);
    current_ios_buffer_.assign(pbase(), pbase() + chunk_size);
  }
  hash_validator_->Update(current_ios_buffer_);

  auto result = upload_session_->UploadChunk(current_ios_buffer_, upload_size);
  if (!result.ok()) {
    // This was an unrecoverable error, time to signal an error.
    return std::move(result).status();
  }
  current_ios_buffer_.clear();
  current_ios_buffer_.reserve(max_buffer_size_);
  current_ios_buffer_.append(trailing);
  auto pbeg = &current_ios_buffer_[0];
  auto pend = pbeg + current_ios_buffer_.size();
  setp(pbeg, pend);
  pbump(static_cast<int>(trailing.size()));

  if (final_chunk) {
    upload_session_.reset();
  }

  // If `result.ok() == false` we never get to this point, so the last response
  // was actually successful, represent that by a HTTP 200 status code.
  last_response_ = HttpResponse{200, std::move(result).value().payload, {}};
  return last_response_;
}

}  // namespace internal
}  // namespace STORAGE_CLIENT_NS
}  // namespace storage
}  // namespace cloud
}  // namespace google
